/* For free support for VSIDE, please visit www.vsdsp-forum.com */

// 3-Button HiFi player (VS1053 only)
// - a reduced version of the full HiFi player code,
// which is runnable on the VS1053 Prototyping Board v2.0.

/// \file hifiplayer.c -- VS1053 Stand-Alone MP3 Player Source Code
/// \author Panu-Kristian Poiksalo, VLSI Solution Oy 2009-2010
/// Software based on the VS1053 Stand-alone player source code by Pasi Ojala
/// Copyright (C) 2005-2010 VLSI Solution Oy Tampere Finland
// See readme.txt.

/// SUPPORTED HARDWARE:
/// This code runs on the VS1053 Prototyping Board V2.0 PCB.

/// Warning: This code is in beta testing phase.
/// Warning: This code inherits a lot of different #ifdef settings from the
///					 VS1053 Stand-Alone player code. Changing these settings is not
///					 recommended and probably most of them do not work without modifications
///					 with the HiFi Player hardware.

// Running:
// - Right click Solution to set Solution-Properties-Debugging-Target Speed to 9600.
// - Remove SD card from the VS1053 Prototyping Board
// - Remove SPIB jumper
// - Switch board power off and on
// - Press the "Run" button in VSIDE.
// Sorry for the long loading time - it's done at 9600 bps so that we never need to change
// the serial port speed to other than 9600 bps, so console messages will be most reliable.

// Define USE_VS3EMU_DEBUG to see messages on Standard Input/Output 
// WARNING: REQUIRES connection to VS3EMU, waits forever if no RS-232 is connected.
#define USE_CONSOLE_DEBUG
// So if you burn the software to EEPROM, be sure to disable console debug first.

// Some of the comments below may be inaccurate because they are from the HiFiplayer, 
// and the comments are not 100% cleaned yet. Basically this is the HiFi player with
// the USB and LCD screen removed and the buttons handling changed to 3 buttons.

// Still apologies for the cryptic code, this descends from some very esoteric stuff.
// The thing is, that although we have a lot more modern and cleaner code in use
// at VLSI today, it's still not ready to publish, and this old code, though it is
// quite cryptic, actually works extremely well. I have tried to comment it a lot for you
// so it would be at least a bit understandable. 
//
// To understand a bit more about the player framework, you can read the
// book "First Steps with VS1053". The basic idea is that MyMain handles reset
// and UserHook is a state machine that handles everything else (SD read, user 
// interface etc). UserHook makes a soft reset at end of song. The "master" of the software
// is any one of the audio decoders in ROM, which calls the UserHook whenever there
// is free CPU time. The UserHook then delivers some more data for the decoder.

// The files are referred to as numbers, not their file name. To find a file with
// specific name, call OpenFile(n) with n = {0,1,2,3,...} and see if the
// char longFileName[] contains the file name you were looking for. 
// Subdirectories are travelled as they are encountered in the FAT records.

// EEPROMMING: Works. Remove the USE_CONSOLE_DEBUG define, rebuild solution and
// prom the EEPROM image with "Project -> Prommer/Flasher Utility". With define
// SKIP_UI_CODE_NOT_AVAILABLE_IN_PROTOTYPING_BOARD, the image size is 8150 bytes,
// just barely smaller than 25LC640 (8192 bytes). If you add code, get a larger
// EEPROM such as 25LC256 or use an SPI FLASH.


/*
	Notes:
	To save code space, the following files are by default excluded from build,
	because the HiFi player with standard settings does not need them:
	asm-clocksdi.s, asm-imafix.s, asm-myrand.s, asm-saveword.s, asm-swap.s,
	asm-swapword.s, asm-writesec.s, asm-writesecswap.s, buts.s, butsrec.s,
	scis.s, scisrec.s, uartasm.s.
									
*/			

#include "defines.h"
#include "gpio.h"

//#define NOCHECKFILETYPE 
#define SHUFFLE_PLAY /*shuffle instead of random +12/19 words*/
#define SKIP_ID3V2		/* skip ID3v2 tags */
#define ENABLE_UI
#define FIX_WAV_DECODETIME	/* SCI_UI: advance decodetime for short WAV's */


#define SKIP_UI_CODE_NOT_AVAILABLE_IN_PROTOTYPING_BOARD							


#include <string.h>
#include "standalone.h"
#include <stdio.h>
#include "mmcfs.h"

s_int16 bassTarget = 0; // target value for BASS register (for smooth changing)
u_int16 bassDelay = 0;

u_int16 GetI6(void);
auto void DecodeTimeClear(void);

#define FAT_LFN_SIZE 100
u_int16 longFileName[FAT_LFN_SIZE];
u_int16 fileSizeW;


// The SPI state machine states in UserHook
typedef enum {
	spi_pause=0,  //Paused.
	spi_file,     //Opening a file
	spi_seek,     //Send Seek (Read) command to SD card
	spi_waitdata, //Wait for data to come from the SD card
	spi_data,     //Read data from the SD card
	spi_zeros,    //End of song, send endFillBytes to the decoder.
	spi_record,   //Recording, not supported in this software
	spi_waitima   //Recording, not supported in this software
} SPISTATE;



SPISTATE spiState, oldSpiState;
u_int16 spiCnt;
u_int16 timeOut;

struct SPIX { /* do not touch! ASM uses the structure */
	s_int32 size;
	u_int32 sector;
	u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
	__y struct FRAGMENT *fragment;
} spix;


register __a1 short SpiDelay(void);
void Delay(u_int32 n){
	while (n--) {
		SpiDelay();
	}
}


#ifdef USE_CONSOLE_DEBUG
#define DebugPrint(a){puts(a);}
void DebugPrintPacked(char *s){
	while (*s>>8) {
		putchar(*s>>8);
		if ((*s&0xff)==0) {
			return;
		}
		putchar(*s++&0xff);
	}	
}
#else
#define DebugPrint(a){}
#define DebugPrintPacked(a){}
#endif

#include <stdlib.h>

#if defined(USE_DEBUG) && !defined(UART_BUFFERED)
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC))
#else
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#endif


//#define HDATCHECK 800 /* 21words: give up after 800/2=400kB ~ 3-6sec */


#ifndef NOCHECKFILETYPE
#define CHECKFILETYPE /* vs1011 29 words? check file type according to short-filename suffix */
#endif/*NOCHECKFILETYPE*/
//#define KEEP_DRIVERS_OFF /* .. until MMC located */
#define WITH_SUBDIRECTORIES /* support subdirectories */

s_int16 hcShift; // sector->byteaddress adjustment for non-SDHC cards.

#ifdef CHECKFILETYPE
#undef CHECKFILETYPE /*temporary undef*/
#define CHECKFILETYPE CheckFileType1053
#endif

#ifdef FATINFO_IN_Y
__y struct FATINFO fatInfo;
#else
struct FATINFO fatInfo;
#endif

//#define MAX_FRAGMENTS 28//16//22
__y struct FRAGMENT fragments[MAX_FRAGMENTS];
u_int16 gFileNum[2];
#include "handledir.inc"

#ifdef HDATCHECK
u_int16 hdatCheck;
#endif

/* Can be done for all formats, because IdleHook is run as the main thread
	 and not in interrupts. Whenever IdleHook is called, there either is
	 not enough stream data anyway, or the decoder is outputting samples
	 and there is not enough space in the audio buffer (thus not reading
	 stream). */
u_int16 __x * __x stream_save;

__y struct {
	u_int16 Data; /* must be first entry! Used by c.s */
	u_int16 Cnt;
	u_int16 Strobe;
	u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
	u_int16 Key;
	s_int16 KeyTime; 
} ui;



#ifdef SHUFFLE_PLAY
u_int16 shuffle_offset = 0;
#endif

int rewmode = 0;
int rewcount = 0;


u_int16 pauseOn = 0;
u_int16 autoPowerOff;





/// This function reads the GPIO keys and changes various player settings
void UserInterface(u_int16 highestAllowedKey) {
	SPISTATE cachedSpiState = spiState;
		
#ifdef ENABLE_UI
			
	// ui.Strobe is 1 about 16 times per second
	if (ui.Strobe) { // PKP: It's time to check the buttons and do UI calculations
		
		ui.Data = ui.Strobe = 0;
		// PKP: -- UI: READ KEYS : 3-button

		ui.Data = 0;
		if (BUTTON_1_PRESSED) {
			ui.Data = PROTOBOARD_SW1;
		} else if (BUTTON_2_PRESSED) {
			ui.Data = PROTOBOARD_SW2;
		} else {
			GPIO_SET_LOW(1 << 0); // Pull GPIO0 (eeprom xcs) low
			if (BUTTON_1_PRESSED) { //Actually button 3 because of diode D6.
				ui.Data = PROTOBOARD_SW3;				
			}
			GPIO_SET_HIGH(1 << 0);		
		}
		
		

		if (ui.Data > highestAllowedKey) {
			ui.Data = 0;
		}		

		
		// PKP: -- UI: KEY FUNCTIONALTY -- // User Interface functions
		
		parametric_x.playSpeed = 1;		 
		
		if (ui.Data == ui.Key) { // PKP: Current Keystate is same as previous Keystate
			if (ui.KeyTime < 8) {
				ui.KeyTime++;
			} else // PKP: keystate has been consistent for 8 ticks - it's a long keypress //
				if (ui.Key	) {

					/* long presses */
					register u_int16 vol = USEX(SCI_VOL); // get a copy of volume setting
					switch(ui.Key) { // Which keypress message is in ui.Key
										
					case PROTOBOARD_SW1: //Intentional fall-through				
					case SW_VOL_UP: //Long press of Volume Up
						if (vol) {
							vol -= 0x0101;
							goto lsetvol;
						}
						break;
						
					case PROTOBOARD_SW2:
					case SW_VOL_DN: //Long press of Volume Down
						//if (vol != (s_int16)0xffffU)
						if ((vol&0xff) < 198){
							vol += 0x0101;
						}
						lsetvol:
						USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
						break;

#ifndef SKIP_UI_CODE_NOT_AVAILABLE_IN_PROTOTYPING_BOARD							
					
					case SW_PREV: // Long press of PREV == Rewind
						if (rewmode == 0) {
							rewmode = 1; //rewind mode (limited support...)
						}
						break;											
						
					case SW_NEXT:	 // Long press of NEXT == Fast forward
						parametric_x.playSpeed = 6;
						break;

#endif

					default: ;
					} //end switch-case											
				}
			
		} else { // PKP: Keystate changed (a key has been pressed or released)
			if (0 == ui.Data) { // All keys are released

				// PKP: Handle end of short keypress (current keystate ui.Data=0)
				if (ui.KeyTime < 8) { /* was pressed < 0.5 seconds */
					register u_int16 vol = USEX(SCI_VOL);

					switch (ui.Key) {
							
					case SW_REC: // Record button short press
						break;
		
#ifndef SKIP_UI_CODE_NOT_AVAILABLE_IN_PROTOTYPING_BOARD							
	
					case SW_VOL_UP: // Volume step up
						if (vol>0x0202) {
							vol -= 0x0303;
							goto lsetvols;
						}
						break;

						
					case SW_VOL_DN: // Volume step down
						if ((vol&0xff) < 194){
							vol += 0x0303;
						}										
						lsetvols:
						USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
						break;
						
		
					case SW_MODE: //Short press of Mode
						// Toggle shuffle play
						USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
						break;
#endif
									
						// Change the song which is playing	
					case PROTOBOARD_SW1: //intentional fall-through
					case PROTOBOARD_SW2:	//intentional fall-through							
					case SW_NEXT: //intentional fall-through
					case SW_PREV:
						{
							register u_int16 song = USEX(SCI_AICTRL0);

							if ((ui.Key == PROTOBOARD_SW1)	
							|| (ui.Key == SW_NEXT)) { // Next song (song + 1)
								song++; /* next file */
							} else
							#ifndef UI_PAUSE_BEFORE_PLAY
							if (USEX(SCI_DECODE_TIME) < 5) //Song played for less than 5 seconds
							#endif
							{
								if (song == 0) {
									song = USEX(SCI_AICTRL1)-1;
								} else { /* previous file */
									song--;
								}
								/* else the same file */
							}
							USEX(SCI_AICTRL0) = song;
							
							/* especially needed for MIDI!! */
							#if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
							USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
							#endif
							cachedSpiState = spi_zeros;
	
							#if defined(ENABLE_LAYER12)
							/* gain=0 for partial last frame (bugs) */
							volume[0] = volume[1] = 0;								
							#endif
						}
						break;
							
						
					case PROTOBOARD_SW3: //Intentional fall-through
					case SW_PLAY:
						/* play pause/continue -- not very good for midi.. */
						pauseOn ^= 1; // Toggle pause
						break;

					
					default: 
						;
					} //end of switch-case
				}//(ui.keyTime was < 8) (Short keypress)
			}
			ui.Key = ui.Data;
			ui.KeyTime = 0;
		}	 
	}		
	spiState = cachedSpiState;
#endif /*ENABLE_UI*/
}





/*
	This state machine function is responsible for doing everything after reset.
*/
void UserHook(void) {
	static SPISTATE cachedSpiState;



	// Smoothly adjust loudness (Bass and Treble).
	// SCI_BASS is changed slowly to "bassTarget" to avoid clicks in sound.
	if (ui.Strobe) {
		if (bassDelay) {
			bassDelay--;
		} else {			
			register s_int16 bassLevel = USEX(SCI_BASS) & 0xf0f0;
			bassLevel |= bassTarget & 0x0f0f;
			
			if ((s_int16)(bassTarget & 0xf000) > (s_int16)(bassLevel & 0xf000)) {
				bassLevel += 0x1000;
			} else if ((s_int16)(bassTarget & 0xf000) < (s_int16)(bassLevel & 0xf000)) {
				bassLevel -= 0x1000;
			}
			if ((bassTarget & 0x00f0) > (bassLevel & 0x00f0)) {
				bassLevel += 0x10;
			} else if ((bassTarget & 0x00f0) < (bassLevel & 0x00f0)) {
				bassLevel -= 0x10;
			}
			USEX(SCI_BASS) = bassLevel;					
		}		 
	}
	// End of bass update functionality




	#ifdef ENABLE_UI
		UserInterface(0xff);
	#endif /*ENABLE_UI*/
	
	cachedSpiState = spiState;	
		
	#ifdef PLAYTIME
		if (USEX(SCI_DECODE_TIME) > PLAYTIME && cachedSpiState != spi_zeros) {
			USEX(SCI_AICTRL0)++;
			#if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
				USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
			#endif
			cachedSpiState = spi_zeros;
		}
	#endif/*PLAYTIME*/
	
		
	///////////  STAND-ALONE PLAYER SPI BUS / SD CARD STATE MACHINE ///////////


	// File found	
	if (cachedSpiState == spi_file) {

		if (0==longFileName[0]) { // No long filename, use 8 chars from short filename
			u_int16 i;
			u_int16 *p = &longFileName[0];
			for (i=0; i<4; i++){
				*p++ = fileName[i] >> 8;
				*p++ = fileName[i] & 0xff;
			}
			*p=0;
		}			 

		DebugPrint((char*)longFileName);

		/* extra byte read if odd file size */
		spix.size = (fatInfo.fileSize+1)/2; /* in words! */
		cachedSpiState = spi_seek;
	setfrag:
		spix.fragment = &fragments[0];
		spix.sector = fragments[0].start /*& 0x7fffffffUL*/;
		spix.fragSize = fragments[0].size; /* *fatInfo.sectorsPerCluster*/
	}
	
	
	

	// State: Seeking for data  or  Pause
	if (cachedSpiState == spi_seek || cachedSpiState == spi_pause) {
		// MMC/SD card sector read is over, SPI bus can be used for other operations
		
		if (pauseOn) {
			if (cachedSpiState != spi_pause) {
				oldSpiState = cachedSpiState;
				cachedSpiState = spi_pause;
				stream_save = stream_wr_pointer;
				stream_wr_pointer = stream_rd_pointer+1;
				
				clockX = 2; //1.0x
				SetRate(hwSampleRate);
			} 
		} else {
			if (cachedSpiState == spi_pause) {
				cachedSpiState = oldSpiState;
				/* restore 'stolen' data */
				stream_wr_pointer = stream_save;
				SetClockFromSpi();		
			}
		}
				
		SpiSendClocks(); /*Raise MMC/SD CS (and send some clocks) */
		USEX(GPIO_ODATA) |= SPI_SD_CS;
		SpiSendClocks();				
	}


	
	// State: seek for data 
	if (cachedSpiState == spi_seek) {
		if (rewmode == 1) {
			s_int32 skip = spix.fragment->size - spix.fragSize;
			if (skip > REWIND_STEP_SIZE)
				skip = REWIND_STEP_SIZE;
			spix.sector -= (u_int16)skip;
			spix.fragSize += (u_int16)skip;
			spix.size += skip*512/2;
			
			rewcount = REWIND_STEP_SIZE / 8;
			rewmode = 2;
		} else if (rewmode == 2) {
			if (--rewcount == 0)
				rewmode = 0;
		}
		
		/* In case we are just trying to find something to play,
			 also read keys once in a while.. */
		ui.Cnt += 64;
		if (ui.Cnt > hwSampleRate/16) {
			ui.Cnt -= hwSampleRate/16;
			ui.Timer++;
			ui.Strobe = 1;
		}

		MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
							 (spix.sector & 0x7fffffffUL) << hcShift);
		cachedSpiState = spi_waitdata;
		spiCnt = 0;
		NextSector();
		
		#ifdef HDATCHECK
			if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
				/* if file not detected at HDATCHECK/2 kB, skip the rest */
				goto end_of_file;
			}
		#endif
	}


	// State: Seek (read) command sent, wait for data from MMC/SD card	
	if (cachedSpiState == spi_waitdata) {
		if (SpiSendReceiveMmc(0xff00,8) == 0xfe) {
			cachedSpiState = spi_data;
			spiCnt = 0;
			timeOut = 0;
		} else {
			if ((s_int16)++spiCnt < 0) {
				goto goreset; //MyReset(); /* timeout */
			}
		}	
	} 

	// State: Input data from MMC/SD card to Stream Buffer	
	else if (cachedSpiState == spi_data) {
		while (spix.size > 0) {
			register u_int16 i, j = 32;
			register u_int16 *wr;
	
			{
				if (StreamDiff() >= STREAM_BUFFER_SZ-40)
					break;

				wr = (void *)stream_wr_pointer;
				if (j > spix.size)
					j = (u_int16)spix.size;
				for (i=0;i<j;i++) {
					*wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
					if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
						wr = stream_buffer;
				}
			}
			stream_wr_pointer = (void *)wr;
			spix.size -= j;//32;
			spiCnt += 32;
			if (spiCnt >= 256) {
				/* 256 words read, get next block in file */
				cachedSpiState = spi_seek; // Next state is to seek new sector
				SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
				break;
			}
			timeOut = 0;
			/* watchdog clearing here doesn't work with midi */
		}	
		
		if ((s_int16)++timeOut == -1) {
			spix.size = 0;
		}		
		if (spix.size < 0x20000L) {
			/* Workaround for WMA resync after stream ended! */
			parametric_x.resync = 0;
		}
		
		
		#ifdef ENABLE_WATCHDOG
			USEX(WDOG_RESET) = WDOG_RESET_VAL;
		#endif

		if (spix.size <= 0) {
			// End of File - Switch to SPI_ZEROS state for "sending" zeros to the
			// audio decoder for resetting the audio decoder.
			end_of_file:
			SpiSendClocks();			
			{
				{  // Figure out which file to play next (SCI register AICTRL0 holds the current file number)
					register int m = (USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK);
					if (m == CTRL3_LOOP_SONG) {
						/* restart the same file immediately */
						cachedSpiState = spi_file;
					} else if (m == CTRL3_PAUSE_AFTER_PLAY) {
						oldSpiState = spi_zeros;
						cachedSpiState = spi_pause;
					} else {
						USEX(SCI_AICTRL0) += 1; /* next file */
						cachedSpiState = spi_zeros;
						if (USEX(SCI_HDAT1) == (('f'<<8) | 'L'))
						USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
					}
				}
			}
		}

		#ifdef NO_WMA /* reject WMA files */
			if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
				USEX(SCI_AICTRL0) += 1; /* next file */
				goto goreset;//MyReset();
			}
		#endif
		#if defined(NO_AAC) && !defined(SCI_UI) && !defined(RECORDER)
			{
				register u_int16 t = USEX(SCI_HDAT1);
				//0x4154(AT) 0x4144(AD) 0x4d34(M4)
				if (t == 0x4154 || t == 0x4144 || t == 0x4d34) {
					USEX(SCI_AICTRL0) += 1; /* next file */
					goto goreset; //MyReset();
				}
			}
		#endif
		

	// State: Send endFillBytes to reset decoders and jump out of state machine to reset.
	} else if (cachedSpiState == spi_zeros) {
			
		parametric_x.resync = 0;		/* Workaround for WMA resync after stream ended! */	
			
		if (StreamDiff() < STREAM_BUFFER_SZ-40) {
			register int i, z = parametric_x.endFillByte;
			register u_int16 *wr = (void *)stream_wr_pointer;
			for (i=0;i<32;i++) {
				*wr = z;
				if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
					wr = stream_buffer;
				} else {
					wr++;
				}
				++spiCnt;
				stream_wr_pointer = (void *)wr;
			}
			/* watchdog clearing here doesn't work with midi */
			timeOut = 0;
		}

		if ((s_int16)++timeOut < 0) {
			goto goreset;
		}
			
		// Time to exit?
		if (USEX(SCI_HDAT1) == 0 || USEX(SCI_HDAT1) == 0x574d || spiCnt >= 256+2048) { 
			/* was left to max 256 by spi_data */
			/* 2048 words instead of bytes */
		goreset:
			/* Discard any unread data, including CRC, before starting again */
			{
				register int i;
				for (i=0; i<256+2; i++) {
					SpiSendReceiveMmc(-1/*0xffff*/,16); //Pump any remaining data from SD card
				}
			}
			SpiSendClocks(); //release chip selects, send dummy SPI clocks
			USEX(DAC_VOL) = 0xaaaa; // Set very low volume
			Delay(10000);
			MyReset(); // ASM reset to Main(). Bye bye.
		}					
		#ifdef ENABLE_WATCHDOG
		goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
		#endif
	} 
	
	// State: any other state.
	else {
		#ifdef ENABLE_WATCHDOG
		wdogack:
		USEX(WDOG_RESET) = WDOG_RESET_VAL;
		#endif
	}
	// END OF STATE MACHINE



				
	#ifndef NO_DREQ_LED
	/* 15 words */
	{
		register int t = 0;
					
		if ((ui.Timer & 0x18) == 0) {
			/* DREQ indicates random play status */
			#ifdef NO_RANDOM
				t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
			#else
				t = USEX(SCI_AICTRL3);	/* lowest bit is global random play */
				#ifdef LOUDNESS_ALWAYS_ON_OFF
					/* indicate random play */
				#else /*LOUDNESS_ALWAYS_ON_OFF*/
											
					if (cachedSpiState
					#ifdef SW3_RANDOM_PLAY
						==
					#else
						!=
					#endif
					spi_pause) {
						t = USEX(SCI_BASS);  /* has lowest bit set if enabled */
					}
				#endif /*else LOUDNESS_ALWAYS_ON_OFF*/
			#endif /*else NO_RANDOM*/
		}
		USEX(SER_DREQ) = t;
	}
	#endif/*!NO_DREQ_LED*/
					
	spiState = cachedSpiState;
			
}
		
	

// Start up the SD card
void InitSD() {
	DebugPrint("Looking for SD card...");

	// Initialize MMC (Inline code saves 16 words)	
	/* wake up MMC! -- If this takes too long, watchdog will bite. */
	{
		static u_int16 retries = 5;
		register s_int16 i, cmd;
		s_int32 d;
			
		// Start the MMC/SD card
	tryagain: 

		ui.Strobe = 1;
		UserInterface(SW1);

		// MMC reset
		USEX(GPIO_DDR) |= (1 << GPIO_PIN_SDCARD_OFF)/*SO, output*/;
		USEX(SER_DREQ) = 0;
		if (++retries>=25) {
			u_int16 gpio0_save = USEX(GPIO_ODATA);
			retries = 0;
			USEX(GPIO_ODATA) = 1 + (1 << GPIO_PIN_SDCARD_OFF); //all 0 except eecs and GPIO OFF
			Delay(100000); //keep SD card switched off for a while
			USEX(GPIO_ODATA) = gpio0_save;
		}
		USEX(GPIO_ODATA) &= ~(1 << GPIO_PIN_SDCARD_OFF); //SD ON
		Delay(10000);			
		for (i=0; i<16; i++) {
			SpiSendClocks();
		}
			

		/* MMC Init, command 0x40 should return 0x01 if all is ok. */
		i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
		for (cmd=0; cmd<512; cmd++) {
			SpiSendReceiveMmc(-1,16);
		}
		
		SpiSendClocks();
		
		if (i != 1)
			goto tryagain;//continue; /* No valid idle response */
			
		cmd = MMC_SEND_OP_COND|0x40;

		if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
				== 1) {
			/* MMC answers: 0x05 SD answers: 0x01 */
			/* Should we read the whole R7 response? */
			//SpiSendReceiveMmc(-1, 32);
			cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
		}

				
		/* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
		//i = 0; /* i is 1 when entered .. but does not make the code shorter*/
		while (1) {
			register int c;

			if (cmd == (0x40|41)) {
				MmcCommand(0x40|55/*CMD55*/,0);
				c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0x40000000UL);
			} else

				c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0);
			if (c == 0)
				break;

			if (++i >= 30000/*< 0*/ || c != 1) {
				goto tryagain; /* Not able to power up mmc */
			}
		}

		// Detect SDHC cards				
		hcShift = 9;
		i = 0x0246;
		/* PRETEC's 1G MMC does not seemd to handle CMD58 here! */
		/* Support HC for SD, but not for MMC! */
		if (cmd == (0x40|41) &&
				MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
			if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
				/* OCR[30]:CCS - card capacity select */
				/* HighCapacity! */
				/* ldc 9,D0 (0x0246) <-> ldc 0,D0 (0x0006) */
				i = 6;
				hcShift = 0;
			}
		}
		// Hack to allow value saving also in VS1003 (IRAM is not cleared at reset)
		WriteToProgramRam((u_int16)ReadDiskSector+2, 0, i); 
				
		/* Set Block Size of 512 bytes -- default for at least HC */
		/* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
		/* Must check return value! (MicroSD restart!) */
		i = MmcCommand(MMC_SET_BLOCKLEN|0x40, 512);
		if (i != 0) {			 
			goto tryagain;
		}
		/* All OK return */

	}
	// The MMC is now initialized. Time to init the filesystem.
}	
		
	
	
	
/// This function is called at startup and also at some other times,
/// for example when the file of the SD card has been played.
/// Its purpose is to initialize the system and start playing.
auto void MyMain(register s_int16 __a1 fullReset) {

	static u_int16 r;
	static u_int16 displayState = 0;
	
	
	if (fullReset) {
		DebugPrint("Power on reset");
		USEX(SCI_VOL) = 0x2020; // Start-up volume
	}

	longFileName[0] = 0;

	USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power (not drivers) */
	USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
	USEX(SCI_DECODE_TIME) = 0;	
	applAddr = ApplAddrRef; //This AIADDR function continually sets higher reference (more volume)
	USEX(SCI_STATUS) |= (1 << SCIST_REFERENCE_SEL);
	
	SpiSendClocks(); // SpiSendClocks releases all chip selects and sends some clocks to SPI bus
	DebugPrint("VS1053 3-Button HiFi Player");

	USEX(SER_DREQ) = 0; /* Led / Aux key scan off*/
	USEX(WDOG_RESET) = WDOG_RESET_VAL;			
	USEX(VS1053_IROM4) &= ~(IROM4_ENABLE); //Enable decoders in IROM4
	// Configure the chip mode and GPIO pins
	USEX(SCI_MODE) = SciModeOr() | (USEX(SCI_MODE) & ~(1<<SCIMB_OUT_OF_WAV));
	USEX(GPIO_ODATA) = SPI_xCS|SPI_CLK|SPI_SD_CS |(1<<LCD_GPIO_PIN_A0)|(1<<LCD_GPIO_PIN_CS); /*use same= -1 word*/
	USEX(GPIO_DDR) = SPI_xCS|SPI_CLK|SPI_SD_CS|(1<<LCD_GPIO_PIN_A0)|(1<<LCD_GPIO_PIN_CS);

	DebugPrint("Now do InitHardware. If it hangs here, check that COM target speed is 9600.");		
	InitHardware(); /* Basic chip initialization, no longer clears memory in vs1053 */
	// Note: resets UART speed to 9600bps.	
	DebugPrint("InitHardware ok");

#ifndef USE_DEBUG
	/* Allow reflashing without MMC/SD inserted. */
	USEX(INT_ENABLE) = (1<<INT_EN_RX);
#endif

	// Common decoder init
	InitLayer3();
	InitCommon();
	InitDecode();

		
	/* If not powerup reset, re-enable analog immediately. */
	if (!fullReset) {
		USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
	}
	applAddr = ApplAddrButRef;


	
	// Initialize SD card and start playing song number SCI_AICTRL0


	 
	/* Allow DREQ and enable interrupts, VERY IMPORTANT! */
	USEX(INT_GLOB_ENA) = 0;
	USEX(INT_GLOB_ENA) = 0;

	InitReadFrame();
	DecodeTimeClear();

	// Init the SD card
	InitSD();

	// Init File System (FAT)
	SpiSendClocks();
	DebugPrint("Scanning File System");		
	memset(stream_buffer, 0, 512/2); // Clear the start of stream buffer (used by SD asm routines)


	r = InitFileSystem();
	/* Detects cards without partition table */
	if (r == 0) { // No error at InitFileSystem
		{
			DebugPrint("Searching for files");
			/* determine number of playable files */
			USEX(SCI_AICTRL1) = OpenFile(0xffffU);
		}				

		if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
			/*shuffle 23 words*/
			USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
		}

		// VS1053: Must set bass enhancer slowly to avoid noise
		// when volume is high
		if (USEX(SCI_BASS)){
			USEX(SCI_BASS) = LOUDNESS_MIN_VALUE;
		}
		bassDelay = 5;

		// Try to open file number SCI_AICTRL0 
		if (OpenFile(USEX(SCI_AICTRL0)) < 0) {
			pauseOn = 0;
			
			/* check for ID3v2 and skip over if found */
			/* An ID3v2 tag can be detected with the following pattern:
				 $49 44 33 yy yy xx zz zz zz zz */
			ReadDiskSector(fragments[0].start);
			if ((GetLong(0) & 0xffffffL) == 0x334449) {
				/* skip in 16kB resolution, assume tags are smaller than 2MB */
				/* 0.5kB resolution now */
				register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
				/* guard against small fragments */
				if (skip >= fragments[0].size) {
					skip = fragments[0].size-1;
				}
				fragments[0].start += skip;
				fragments[0].size	 -= skip;
				fatInfo.fileSize -= (u_int32)skip * 512;
			}
	

			if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_PAUSE_BEFORE_PLAY) {
				USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY;
				oldSpiState = spi_file;
				spiState = spi_pause;
				stream_save = stream_wr_pointer;
			} else {
				spiState = spi_file;
			}
			hdatCheck = 0;

		} else {
		gofirst:
			USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
			MyReset(); //Restart
		}
	} else {
		/* Initialize again */
		goto gofirst; //MyReset(); //Restart
	}

		
	USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
	USEX(SCI_DECODE_TIME) = 0;

	parametric_x.resync = 32767;

	stream_rd_pointer = stream_wr_pointer = stream_buffer;
	stream_rd_odd = 0;
			
	// All is now ready for playing. UserHook will control everything from here.		
	// Exit to Main loop for MP3. Other codecs DO break out from this loop.
	// Warning: Do not edit this loop. Functionality must be in the hooks.
	while (1) {
		if (!ReadFrame()) {
			play_frame();
		} else {
			InitReadFrame();
		}
	}
}